"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.GetFetchDataExecutor = exports.GetDeployExecutor = exports.GetRefreshExecutor = exports.GetRunStepExecutor = exports.WrapSQLWithExplain = exports.GetExecutorCBForRunContext = exports.ERunContext = void 0;
const Runner = __importStar(require("./Runner"));
const RunnerOperations = __importStar(require("./RunnerOperations"));
const MacroOperations = __importStar(require("./MacroOperations"));
const CommonOperations = __importStar(require("./CommonOperations"));
const Templates = __importStar(require("./Templates"));
const Workspaces = __importStar(require("./Workspaces"));
const MetadataHydration = __importStar(require("./MetadataHydration"));
const State = __importStar(require("./State"));
const Logging = __importStar(require("./Logging"));
const StepTypes = __importStar(require("./StepTypes"));
const ErrorOperations = __importStar(require("./ErrorOperations"));
const TestingOperations = __importStar(require("./TestingOperations"));
const MetadataOperations = __importStar(require("./MetadataOperations"));
const Common = __importStar(require("./Common"));
/*
 *   STAGE EXECUTORS
 *    - these are used to encapsulate the logic for each different run step instance:
 *     - deploy
 *     - refreshes
 *     - executing a node's script against snowflake
 *     - pulling data from snowflake
 */
var ERunContext;
(function (ERunContext) {
    ERunContext["getData"] = "getData";
    ERunContext["runDeploy"] = "runDeploy";
    ERunContext["runRefresh"] = "runRefresh";
    ERunContext["runStep"] = "runStep";
})(ERunContext = exports.ERunContext || (exports.ERunContext = {}));
const GetExecutorCBForRunContext = (runContext) => {
    const executorCBs = {
        [ERunContext.runStep]: exports.GetRunStepExecutor,
        [ERunContext.runDeploy]: exports.GetDeployExecutor,
        [ERunContext.runRefresh]: exports.GetRefreshExecutor,
        [ERunContext.getData]: exports.GetFetchDataExecutor
    };
    return executorCBs[runContext];
};
exports.GetExecutorCBForRunContext = GetExecutorCBForRunContext;
const WrapSQLWithExplain = (SQL) => {
    return `EXPLAIN USING text ${SQL}`;
};
exports.WrapSQLWithExplain = WrapSQLWithExplain;
const deployLogger = Logging.GetLogger(Logging.LoggingArea.Deploy);
/**
 * These functions will always pass standard hydratedMetadata to RunStepExecutor:
 *
 * 1. RunUpdateStep(i.e. validate run/run):
 *      templateString used is the run script for the given node's template
 *      and references standard hydrated metadata shape
 *
 * 2. RunStepPreview (i.e. validate select):
 *      templateString used is a canned, simple Snowflake script used to validate
 *      and refrences standard hydrated metadata shape
 *
 * These functions may pass standard or advanced hydratedMetadata to RunStepExecutor
 * depending on the situation:
 *
 * 1. RunRenderAndValidate (i.e. template editor render):
 *      will pass advanced hydratedMetadata when rendering the create script while
 *      editing/viewing a template with advanced deploy strategy - this is dependent
 *      on the template being viewed, node the node selected in the viewer. Run script
 *      will always pass in standard hydratedMetadata (as of now run script doesn't care
 *      about before/after state of metadata as the create script would in an advanced deploy)
 *
 * 2. RunProvisionTable (i.e. validate create/create):
 *      will pass advanced hydratedMetadata containing only the 'desiredState' since
 *      the node is being created, while 'currentState' will be undefined
 *
 * 3. PlayTaskThunk (i.e. validate run all/run all/validate create all/create all):
 *      When running all or validating run all, PTT will provide standard hydratedMetadata.
 *      When creating all or validating create all, PTT provide either standard hydratedMetadata
 *      for standard deploy nodes, or hydratedMetadata with only 'desiredState' since
 *      the node is being created - 'currentState' will be undefined
 *
 * 4. during deploy - does not call RunStepExecutor:
 *      deploy uses a separate and distinct code path.
 *      hydrated metadata for standard and advanced deploy strategies are
 *      composed in the same workspace edits creation function(s)
 *
 */
class RunStepExecutor {
    constructor(templateString, stepID, currentState, specifiedIndex, initializeCB, fetchDataOnSuccess, RunSQL, isPreview, hydratedMetadata) {
        this.initialize = () => {
            return this.initializeCB();
        };
        this.finalize = () => {
            return Promise.resolve();
        };
        this.getStepExecutions = () => {
            return new Promise((resolve, reject) => {
                const physicalLocations = Workspaces.getCurrentWorkspaceMappings(this.currentState);
                const allMacrosDictionary = Object.assign(Object.assign({}, this.currentState.macros), MacroOperations.refMacroDefinitions);
                const workspaceID = Workspaces.getCurrentWorkspaceID(this.currentState);
                const runTimeParameters = Workspaces.GetRunTimeParametersFromWorkspacesSafe(this.currentState.workspaces, workspaceID);
                const templateString = this.templateString;
                const userContext = CommonOperations.GetUserContextByUserAuthType(this.currentState, Workspaces.getCurrentWorkspace(this.currentState), null);
                const macrosString = MacroOperations.getAllMacros(allMacrosDictionary);
                // render the templates
                Runner.RunRenderSQL(templateString, userContext, this.hydratedMetadata, macrosString)
                    .then((templateRunResult) => {
                    // parse the templates
                    const sqlStageResult = Templates.getParsedSQLStagesFromString(templateRunResult.runDetails.compiledSQL);
                    const hydratedMetadata = Object.assign(Object.assign({}, RunnerOperations.GetBaseMetadata(this.currentState.steps[this.stepID], physicalLocations)), MetadataHydration.hydrateMetadata(this.nodeMetadata, this.stepID, this.currentState.steps, this.currentState.stepTypesYaml, this.currentState.workspaces, this.specifiedIndex, runTimeParameters));
                    const stepExecutions = sqlStageResult.parsedSQLStages.map((parsedSQLStage, index) => {
                        let templateString;
                        if (this.isPreview) {
                            templateString = (0, exports.WrapSQLWithExplain)(parsedSQLStage.code);
                        }
                        else {
                            templateString = parsedSQLStage.code;
                        }
                        const workspaceID = Workspaces.getCurrentWorkspace(this.currentState);
                        const templateParams = {
                            userContext: CommonOperations.GetUserContextByUserAuthType(this.currentState, workspaceID, null),
                            templateString: templateString,
                            nodeMetadata: hydratedMetadata,
                            macrosString: MacroOperations.getAllMacros(allMacrosDictionary)
                        };
                        return {
                            execute: (dispatch) => {
                                return new Promise((resolve, reject) => {
                                    var _a, _b;
                                    // validate any required fields - throw an error if a required field is missing
                                    const stepTypeConfig = (_b = (_a = StepTypes.getNodeSpecSafe(this.currentState.stepTypesParsed, this.nodeMetadata.sqlType)) === null || _a === void 0 ? void 0 : _a.config) !== null && _b !== void 0 ? _b : null;
                                    if (stepTypeConfig) {
                                        const requiredFieldsExist = Templates.ValidateRequiredConfigFlags(this.nodeMetadata, stepTypeConfig);
                                        if (!requiredFieldsExist) {
                                            return reject({
                                                error: {
                                                    errorString: ErrorOperations.ERunStepError.missingAttributesError
                                                }
                                            });
                                        }
                                    }
                                    const startTime = new Date();
                                    const newQueryResult = {
                                        name: parsedSQLStage.name,
                                        isRunning: true
                                    };
                                    dispatch(State.getUpdateQueryResultSequenceAction(this.stepID, newQueryResult));
                                    return this.RunSQL(templateParams)
                                        .then((queryResult) => {
                                        const queryResultForUpdate = Object.assign(Object.assign({}, queryResult), { name: parsedSQLStage.name, type: parsedSQLStage.type });
                                        // if running a test and it fails
                                        if (TestingOperations.IsTestQueryResult(queryResultForUpdate) && TestingOperations.IsFailedTestQueryResult(queryResultForUpdate)) {
                                            queryResultForUpdate.success = false;
                                            queryResultForUpdate.error = { errorString: "Failed Test", errorDetail: "Failed Test" };
                                            // if failed test will not continue on failure - as of now always a Node test (cof not option for col tests)
                                            if (parsedSQLStage.continueOnFailure === Templates.EQueryResultContinueOnFailure.false) {
                                                // update sql error message
                                                const cofErrorString = `Test failed with "continue on failure" set to false`;
                                                queryResultForUpdate.error = { errorString: cofErrorString, errorDetail: cofErrorString };
                                                // update the queryResult
                                                dispatch(State.getUpdateQueryResultSequenceIndexAction(this.stepID, queryResultForUpdate, index));
                                                // end run
                                                return reject({ error: { errorString: cofErrorString, errorDetail: cofErrorString } });
                                            }
                                        }
                                        dispatch(State.getUpdateQueryResultSequenceIndexAction(this.stepID, queryResultForUpdate, index));
                                        resolve(queryResult);
                                    })
                                        .catch(error => {
                                        const formattedError = Runner.HandleError(error, startTime, templateString); //this is where we can add back in the sql to run?
                                        dispatch(State.getUpdateQueryResultSequenceIndexAction(this.stepID, Object.assign(Object.assign({}, formattedError), { name: parsedSQLStage.name }), index));
                                        reject(error);
                                    });
                                });
                            }
                        };
                    });
                    let getFetchDataExecutionPromise;
                    if (this.fetchDataOnSuccess && this.templateString.length) {
                        const dummyExecutor = (params) => {
                            return Promise.resolve(RunnerOperations.CreateBaseIQueryResult());
                        };
                        const getDataExecutor = (0, exports.GetFetchDataExecutor)(this.stepID, null, dummyExecutor);
                        getFetchDataExecutionPromise = () => getDataExecutor.getStepExecutions().then(getDataExecutions => {
                            // getDataExecutor.getStepExecutions will always resolve with an array that has a length of 1
                            stepExecutions.push(getDataExecutions[0]);
                        });
                    }
                    else {
                        getFetchDataExecutionPromise = () => Promise.resolve();
                    }
                    getFetchDataExecutionPromise().then(() => {
                        return resolve(stepExecutions);
                    });
                }).catch(error => {
                    // SQL rendering error
                    reject(error);
                });
            });
        };
        this.templateString = templateString;
        this.stepID = stepID;
        this.currentState = currentState;
        this.specifiedIndex = specifiedIndex;
        this.nodeMetadata = this.currentState.steps[stepID].operation;
        this.initializeCB = initializeCB;
        this.fetchDataOnSuccess = fetchDataOnSuccess;
        this.hydratedMetadata = hydratedMetadata;
        this.RunSQL = RunSQL;
        this.isPreview = isPreview;
    }
}
class DeployExecutor {
    constructor(workspaceEdits, templateParams, stepID, initializeRunCB, firebaseUpdateCB, currentState, RunSQL) {
        this.initialize = () => {
            return this.initializeRunCB(this.stepID);
        };
        this.finalize = () => {
            return this.firebaseUpdateCB(this.stepID);
        };
        this.getStepExecutions = () => {
            return new Promise((resolve, reject) => {
                const renderSQLPromiseArray = [];
                const runExecuteCBAndUpdateFBWithQueryResult = (dispatch, exeCB, name, queryIndex, startTime, templateParams) => {
                    //every templateParams must have a templateString associated with it
                    Common.assert(Logging.LoggingArea.Runner, !!templateParams.hasOwnProperty("templateString"), "Template string exists");
                    return new Promise((resolve, reject) => {
                        this.firebaseUpdateCB(this.stepID)
                            .then(exeCB)
                            .then((queryResult) => {
                            // sql baked into queryResult
                            dispatch(State.getUpdateQueryResultSequenceIndexAction(this.stepID, Object.assign(Object.assign({}, queryResult), { name }), queryIndex));
                            return this.firebaseUpdateCB(this.stepID).then(() => {
                                resolve(queryResult);
                            });
                        })
                            .catch(error => {
                            const formattedError = Runner.HandleError(error, startTime);
                            dispatch(State.getUpdateQueryResultSequenceIndexAction(this.stepID, Object.assign(Object.assign({}, formattedError), { sql: templateParams.templateString, name }), queryIndex));
                            reject(error);
                        });
                    });
                };
                this.workspaceEdits.forEach((workspaceEdit) => {
                    const nodeMetadata = Object.assign(Object.assign(Object.assign({}, this.templateParams.nodeMetadata), workspaceEdit.dataForNodeMetadata), { parameters: this.templateParams.nodeMetadata.parameters });
                    renderSQLPromiseArray.push(RunnerOperations.RunRenderSQL(workspaceEdit.sql, this.templateParams.userContext, nodeMetadata, this.templateParams.macrosString));
                });
                // get parsedSQLStages for each workspace edit
                Promise.all(renderSQLPromiseArray)
                    .then(allTemplateRunResults => {
                    const stepExecutions = [];
                    allTemplateRunResults.forEach((templateRunResult, templateRunResultIndex) => {
                        let stepExecutionCounter = templateRunResultIndex;
                        const startTime = new Date();
                        const parsedSQLStagesResult = Templates.getParsedSQLStagesFromString(templateRunResult.runDetails.compiledSQL);
                        const hasSQLStagesToExecute = !!parsedSQLStagesResult.parsedSQLStages.length;
                        // source tables that are being removed have no sql stages
                        if (!hasSQLStagesToExecute) {
                            stepExecutions.push({
                                execute: (dispatch) => {
                                    return new Promise((resolve, reject) => {
                                        const executionCb = () => Promise.resolve(RunnerOperations.GetEmptyQueryResult(templateRunResult, startTime));
                                        resolve(runExecuteCBAndUpdateFBWithQueryResult(dispatch, executionCb, "Deleting metadata for source input", 0, startTime, Object.assign(Object.assign({}, this.templateParams), { renderTemplateOnly: true })));
                                    });
                                }
                            });
                        }
                        else {
                            parsedSQLStagesResult.parsedSQLStages.forEach((stage) => {
                                stepExecutions.push({
                                    execute: (dispatch) => {
                                        return new Promise((resolve, reject) => {
                                            const thisStepExecutionCB = () => this.RunSQL(Object.assign(Object.assign({}, this.templateParams), { templateString: stage.code }));
                                            resolve(runExecuteCBAndUpdateFBWithQueryResult(dispatch, thisStepExecutionCB, stage.name, stepExecutionCounter, startTime, Object.assign(Object.assign({}, this.templateParams), { templateString: stage.code })));
                                            stepExecutionCounter++;
                                        });
                                    }
                                });
                            });
                        }
                    });
                    resolve(stepExecutions);
                })
                    .catch(error => {
                    reject(error);
                });
            });
        };
        this.workspaceEdits = workspaceEdits;
        this.templateParams = templateParams;
        this.stepID = stepID;
        this.initializeRunCB = initializeRunCB;
        this.firebaseUpdateCB = firebaseUpdateCB;
        this.currentState = currentState;
        this.RunSQL = RunSQL;
    }
}
class RunRefreshExecutor {
    constructor(templateString, stepID, currentState, specifiedIndex, isPreview, firebaseUpdateCB, initializeRunCB, RunSQL) {
        this.initialize = () => this.initializeRunCB(this.stepID);
        this.finalize = () => {
            return this.firebaseUpdateCB(this.stepID);
        };
        this.getStepExecutions = () => {
            const getMetadataForPlayTaskThunk = () => {
                const runTimeParameters = Workspaces.GetRunTimeParametersFromWorkspacesSafe(this.currentState.workspaces, this.currentState.workspaces.currentWorkspace);
                return MetadataOperations.GetHydratedMetadataWithNodeMetadata(this.stepID, this.currentState, RunnerOperations.OperationTypes.run, this.specifiedIndex, null, runTimeParameters);
            };
            const hydratedMetadata = getMetadataForPlayTaskThunk();
            // this just wraps a normal step execution with firebase updates
            return new RunStepExecutor(this.templateString, this.stepID, this.currentState, this.specifiedIndex, Promise.resolve(), false, this.RunSQL, this.isPreview, hydratedMetadata)
                .getStepExecutions()
                .then(executions => {
                return executions.map((executor, executionIndex) => {
                    return {
                        execute: (dispatch) => {
                            return new Promise((resolve, reject) => {
                                const startTime = new Date();
                                this.firebaseUpdateCB(this.stepID)
                                    .then(() => executor.execute(dispatch))
                                    .then(queryResult => {
                                    this.firebaseUpdateCB(this.stepID)
                                        .then(() => {
                                        resolve(queryResult);
                                    })
                                        .catch(error => {
                                        reject(error);
                                    });
                                })
                                    .catch(error => {
                                    deployLogger.error(`Failure on refresh execution, ${error}`);
                                    const formattedError = Runner.HandleError(error, startTime);
                                    dispatch(State.getUpdateQueryResultSequenceIndexAction(this.stepID, Object.assign(Object.assign({}, formattedError), { name: `Run ${this.nodeMetadata.name}` }), executionIndex));
                                    reject(error);
                                });
                            });
                        }
                    };
                });
            });
        };
        this.templateString = templateString;
        this.stepID = stepID;
        this.currentState = currentState;
        this.specifiedIndex = specifiedIndex;
        this.nodeMetadata = this.currentState.steps[stepID].operation;
        this.isPreview = isPreview;
        this.firebaseUpdateCB = firebaseUpdateCB;
        this.initializeRunCB = initializeRunCB;
        this.RunSQL = RunSQL;
    }
}
//fix later CD-2073
class GetDataExecutor {
    constructor(stepID, token, RunSQL) {
        this.initialize = () => Promise.resolve();
        this.finalize = () => Promise.resolve();
        this.getStepExecutions = () => {
            return Promise.resolve([{
                    execute: (dispatch) => dispatch(Runner.RunGetRows(this.stepID, this.token))
                        .then((queryResult) => {
                        dispatch(State.changeUIState("sqlTabActiveIndex", "Data"));
                        dispatch(State.newDataResultAction(this.stepID, queryResult));
                        dispatch(State.getUpdateQueryResultSequenceAction(this.stepID, Object.assign(Object.assign({}, queryResult), { name: "View Data" })));
                        return queryResult;
                    }).catch(error => {
                        throw error;
                    })
                }]);
        };
        this.stepID = stepID;
        this.token = token;
        this.RunSQL = RunSQL;
    }
}
const GetRunStepExecutor = (templateString, currentState, specifiedIndex, stepID, initializeCB, fetchDataOnSuccess, RunSQL, isPreview, hydratedMetadata) => {
    return new RunStepExecutor(templateString, stepID, currentState, specifiedIndex, initializeCB, fetchDataOnSuccess, RunSQL, isPreview, hydratedMetadata);
};
exports.GetRunStepExecutor = GetRunStepExecutor;
const GetRefreshExecutor = (templateString, currentState, specifiedIndex, stepID, isPreview, firebaseUpdateCB, initializeRunCB, RunSQL) => {
    return new RunRefreshExecutor(templateString, stepID, currentState, specifiedIndex, isPreview, firebaseUpdateCB, initializeRunCB, RunSQL);
};
exports.GetRefreshExecutor = GetRefreshExecutor;
const GetDeployExecutor = (workspaceEdits, templateParams, stepID, initializeRunCB, firebaseUpdateCB, currentState, RunSQL) => {
    return new DeployExecutor(workspaceEdits, templateParams, stepID, initializeRunCB, firebaseUpdateCB, currentState, RunSQL);
};
exports.GetDeployExecutor = GetDeployExecutor;
const GetFetchDataExecutor = (stepID, token, RunSQL) => {
    return new GetDataExecutor(stepID, token, RunSQL);
};
exports.GetFetchDataExecutor = GetFetchDataExecutor;
//# sourceMappingURL=RunStepHelpers.js.map